@@ -3,6 +3,7 @@
 from packet.Vlan import Vlan
 from .baseapi import BaseAPI
 from .baseapi import Error as PacketError
+from .baseapi import ResponseError
 from .Batch import Batch
 from .Plan import Plan
 from .Device import Device
@@ -14,6 +15,7 @@ from .Volume import Volume
 from .BGPConfig import BGPConfig
 from .BGPSession import BGPSession
 from .IPAddress import IPAddress
+from .HardwareReservation import HardwareReservation
 from .Snapshot import Snapshot
 from .Organization import Organization
 from .Email import Email
@@ -73,6 +75,20 @@ class Manager(BaseAPI):
         data = self.call_api("projects", type="POST", params=params)
         return Project(data, self)
 
+    def list_hardware_reservations(self, project_id, params={}):
+        data = self.call_api(
+            "projects/%s/hardware-reservations" % project_id, params=params
+        )
+        hardware_reservations = list()
+        for jsoned in data["hardware_reservations"]:
+            hardware_reservation = HardwareReservation(jsoned, self)
+            hardware_reservations.append(hardware_reservation)
+        return hardware_reservations
+
+    def get_hardware_reservation(self, hardware_reservation_id):
+        data = self.call_api("hardware-reservations/%s" % hardware_reservation_id)
+        return HardwareReservation(data, self)
+
     def list_devices(self, project_id, params={}):
         data = self.call_api("projects/%s/devices" % project_id, params=params)
         devices = list()
@@ -121,6 +137,8 @@ class Manager(BaseAPI):
         user_ssh_keys=[],
         userdata="",
         hardware_reservation_id="",
+        storage={},
+        customdata={},
     ):
 
         params = {
@@ -141,6 +159,10 @@ class Manager(BaseAPI):
 
         if hardware_reservation_id != "":
             params["hardware_reservation_id"] = hardware_reservation_id
+        if storage:
+            params["storage"] = storage
+        if customdata:
+            params["customdata"] = customdata
         if ipxe_script_url != "":
             params["always_pxe"] = always_pxe
             params["ipxe_script_url"] = ipxe_script_url
@@ -175,6 +197,48 @@ class Manager(BaseAPI):
         data = self.call_api("ssh-keys", type="POST", params=params)
         return SSHKey(data, self)
 
+    def create_project_ssh_key(self, project_id, label, public_key):
+        """
+        Successfully creating an SSH key with a Project API Token results
+        in a 404 from the API. If we get a 404, we try the request again.
+
+        If the request actually failed with a 404, we will get another 404
+        which we raise.
+
+        If the request actually succeeded, we will get a 422. In this case,
+        we will try to list all the keys and find the SSHKey we just
+        received.
+
+        Customer Report Reference: TUVD-0107-UIKB
+        """
+
+        def issue_req():
+            try:
+                params = {"key": public_key, "label": label}
+                data = self.call_api(
+                    "projects/%s/ssh-keys" % project_id, type="POST", params=params
+                )
+                return SSHKey(data, self)
+            except ResponseError as e:
+                if e.response.status_code == 422:
+                    # Try to pluck the SSH key from the listing API
+                    keys = [
+                        key
+                        for key in self.list_ssh_keys()
+                        if key.key.strip() == public_key.strip()
+                    ]
+                    if len(keys) == 1:
+                        return keys.pop()
+                raise
+
+        try:
+            return issue_req()
+        except ResponseError as e:
+            if e.response.status_code == 404:
+                return issue_req()
+            else:
+                raise
+
     def list_volumes(self, project_id, params={}):
         params["include"] = "facility,attachments.device"
         data = self.call_api("projects/%s/storage" % project_id, params=params)
@@ -247,8 +311,8 @@ class Manager(BaseAPI):
             )
 
         try:
-            self.call_api("/capacity", "POST", params)
-            return True
+            data = self.call_api("/capacity", "POST", params)
+            return all(s["available"] for s in data["servers"])
         except PacketError as e:  # pragma: no cover
             if e.args[0] == "Error 503: Service Unavailable":
                 return False
@@ -510,12 +574,15 @@ class Manager(BaseAPI):
 
         return vlans
 
-    def create_vlan(self, project_id, facility, vxlan=None, vlan=None):
+    def create_vlan(
+        self, project_id, facility, vxlan=None, vlan=None, description=None
+    ):
         params = {
             "project_id": project_id,
             "facility": facility,
             "vxlan": vxlan,
             "vlan": vlan,
+            "description": description,
         }
         data = self.call_api(
             "projects/%s/virtual-networks" % project_id, type="POST", params=params
